from ._low_index import *

# Version 1.1 was used for the cython implemention
# Version 1.2 (and later) is used for the C++/pybind11 implementation
__version__ = '1.2.1'

def version():
    return __version__



import os, platform

def cpu_info(): 
    if platform.system() == 'Windows': 
        return platform.processor() 
    elif platform.system() == 'Darwin': 
        command = '/usr/sbin/sysctl -n machdep.cpu.brand_string' 
        return os.popen(command).read().strip() 
    elif platform.system() == 'Linux':
        with open('/proc/cpuinfo') as input_file:
            info = input_file.read().split('\n')
            for line in info:
                parts = line.split(':')
                if parts[0].strip() == 'model name':
                    return parts[1].strip()
    return 'Unknown CPU'

def expand_relator(word):
    syllables = []
    for letter in word:
        if letter.isupper():
            syllables.append('%s^-1'%letter.lower())
        else:
            syllables.append(letter)
    return '*'.join(syllables)



import os
import sys
import time
import subprocess
from tempfile import NamedTemporaryFile
import low_index
from low_index import permutation_reps
from low_index import benchmark_util

use_low_level = False
low_level_multi_threaded = False

examples = [
    {
        'group' : 'Fundamental group of K11n34',
        'rank': 3,
        'short relators': ["aaBcbbcAc"],
        'long relators': ["aacAbCBBaCAAbbcBc"],
        'index': 7,
    },
    {
        'group' : 'Fundamental group of K11n34',
        'rank': 3,
        'short relators': ["aaBcbbcAc"],
        'long relators':  ["aacAbCBBaCAAbbcBc"],
        'index': 8,
    },
    {
        'group' : 'Fundamental group of K15n12345',
        'rank': 3,
        'short relators': ["aBcACAcb"],
        'long relators': ["aBaCacBAcAbaBabaCAcAbaBaCacBAcAbaBabCAcAbABaCabABAbABaCabCAcAb"],
        'index': 7,
    },
    {
        'group' : 'Fundamental group of K15n12345',
        'rank': 3,
        'short relators': ["aBcACAcb"],
        'long relators':["aBaCacBAcAbaBabaCAcAbaBaCacBAcAbaBabCAcAbABaCabABAbABaCabCAcAb"],
        'index': 8,
    },
    {
        'group' : 'Fundamental group of o9_15405',
        'rank': 2,
        'short relators': ["aaaaabbbaabbbaaaaabbbaabbbaaaaaBBBBBBBB"],
        'long relators': [],
        'index': 9,
    },
    {
        'group' : 'Fundamental group of o9_15405',
        'rank': 2,
        'short relators': [],
        'long relators': ["aaaaabbbaabbbaaaaabbbaabbbaaaaaBBBBBBBB"],
        'index': 9,
    },
    {
        'group': 'K5_13(3,0)',
        'rank': 2,
        'short relators': ['bbb', 'abABaaBAbabABaaBAbabABAb'],
        'long relators': [],
        'index': 15
    },
    {
        'group': 'K15n145097(0,1)',
        'rank': 3,
        'short relators': ['aBCBabaBAbcbAccAbC'],  # Fastest was just one short rel
        'long relators': ['aBAbcbAcbAcAbAcAbAcAbCaBC',
                          'aBCBabABaaaaaaaaaaaaaaaaBAbcbAccbAcbAcAC'],
        'index': 7,
    },
    {
        'group': 'DodecahedralOrientableClosedCensus[3]',
        'rank': 4,
        'short relators': ['abbDDabacdc',
                           'acdAcDc',
                           'aBacddBBcD',
                           'acddCdCbAddB'],
        'long relators': [],
        'index': 6,
    },
    {
        'group': 'DodecahedralOrientableClosedCensus[8]',
        'rank': 4,
        'short relators': ['adADAddAbdbdbadaDD',
                           'adaDDbcccDAddAbdb',
                           'adaDDbccACACdaDDadCCBdAbcBdAbcBD',
                           'acacaDAddADcacaCCBdAbcBdAbcBdAbc'],
        'long relators': [],  # It didn't seem to help to make some rels long.
        'index': 5,
    },
    {
        'group': 'L14a26995(2,0)(2,0)(2,0)',
        'rank': 4,
        'short relators': ['bb',
                           'aa',
                           'cdcd',
                           'aCABacAddcDbABdCDDaCAbac',
                           'aBdcDbABddcDbaBdCDbABacAb',
                           'aCAbacACBDCbcaCAbacACBcdbcaCABacACBc'],
        'long relators': [], # Making rels long *slowed things down*.
        'index': 7,
    },
    {
        'group': 'ocube06_06795',
        'rank': 6,
        'short relators': ['bdbEED'],
        'long relators': ['aeecbcAfcbcF',
                          'bcedeeccDBec',
                          'aeecfedABaDBeFbcAb',
                          'aeecaeecfeBeFbcAbcAb'],
        'index': 4,

    },
    {
        'group': 'v3448~irr(1,0)(1,0)(1,0)',
        'spec': 'XLLvLPLvwLLvwzAzPvPwMLLzAPQAQQQQQkcdjhijoqnswwuBztxGIEAORLTKGPHUOHNSTSWLWRVQOVPQUVUWhssovshsksxtuxphrreckdlpqxreuaxejknrawarjnsjvrncbd_CbbadEBbBDab(1,0)(1,0)(1,0)',
        'rank': 5,
        'short relators': ['aaaabccccDDb'],
        'long relators': ['aaabcECBAAABAAABEEECBAAABaaaabcedAbaaabceCBAAABaaaabcbaaabcDbb',
            'aaabcdAbaaabceeeedAbaaabceCBAAABBddcDAAAABdAAAABEEECBAAABaaaabcbaaabcDbb',
            'aaabcdaaabcedAbaaabceeeebaaaaDbaaaadCDCBAAABCBAAAAbaaabcECBAAABaDEbaaabaaabcedAbaaabceeeebaaaadCCCCb',
            'aaabcdAbaaabceeeedAbaaabceCBAAABaaaabcbaaabcdaaabcedAbaaabceeeebaaaaDbaaaadCaaabcedAbaaabceeeebaaaaDbaaaadCaaabcedAbaaabceeeebaaaadCCCCb'],
        'index': 4,
    },
    {
        'group' : 'Symmetric Group S7',
        'rank' : 2,
        'short relators': [
            "aaaaaaa",
            "bb",
            "abababababab",
            "AbabAbabAbab",
            "AAbaabAAbaab",
            "AAAbaaabAAAbaaab"],
        'long relators': [],
        'index': 35,
    },
    # This example must go at the end to avoid an infinite loop when
    # using Magma for the benchmark.
    {
        'group' : 'Modular Group',
        'rank': 2,
        'short relators': ["aa", "bbb"],
        'long relators': [],
        'index': 25,
        'gap skip': True,
    },
]


def run_example(ex, num_threads=0):
    return len(
        permutation_reps(
            rank = ex['rank'],
            short_relators = ex['short relators'],
            long_relators = ex['long relators'],
            max_degree = ex['index'],
            strategy = ex.get('strategy', 'spin_short'),
            num_threads = num_threads))

def run_example_low_level(ex):
    short_relators = low_index.spin_short(
        [ low_index.parse_word(
            rank = ex['rank'], word = relator)
          for relator in ex['short relators'] ],
        max_degree = ex['index'])
    long_relators = [
        low_index.parse_word(
            rank = ex['rank'], word = relator)
          for relator in ex['long relators'] ]
    if low_level_multi_threaded:
        tree = low_index.SimsTreeMultiThreaded(
            rank = ex['rank'],
            max_degree = ex['index'],
            short_relators = short_relators,
            long_relators = long_relators,
            num_threads = low_index.hardware_concurrency())
    else:
        tree = low_index.SimsTree(
            rank = ex['rank'],
            max_degree = ex['index'],
            short_relators = short_relators,
            long_relators = long_relators)
    return len(tree.list())

def run(ex, num_threads = 0):
    print('%s; index = %d.'%(ex['group'], ex['index']))
    start = time.time()
    if use_low_level:
        n = run_example_low_level(ex)
    else:
        n = run_example(ex, num_threads = num_threads)
    elapsed = time.time() - start
    print('%d subgroups' % n)
    print('%.3fs'%elapsed)
    sys.stdout.flush()

def run_regina(ex):
    print('%s; index = %d.'%(ex['group'], ex['index']))
    if ex['index'] > 7:
        print("Skipping because regina requires index <= 7")
        sys.stdout.flush()
        return
    start = time.time()
    G = regina.GroupPresentation(
        ex['rank'], ex['short relators'] + ex['long relators'])
    n = 1
    for d in range(2, ex['index'] + 1):
        n += len(G.enumerateCovers(d))
    elapsed = time.time() - start
    print('%d subgroups' % n)
    print('%.3fs'%elapsed)
    sys.stdout.flush()

def translate_to_gap(ex, output):
    output.write('info := "%s; index=%d";\n'%(ex['group'], ex['index']))
    all_relators = ex['short relators'] + ex['long relators']
    gap_relators = [benchmark_util.expand_relator(r) for r in all_relators]
    letters = 'abcdefghijklmnopqrstuvwxyz'
    generators = letters[:ex['rank']]
    output.write('F := FreeGroup(')
    output.write(', '.join(['"%s"'%g for g in generators]))
    output.write(');\n')
    for n, gen in enumerate(generators):
        output.write('%s := F.%d;\n'%(gen, n + 1))
    output.write('G := F / [\n')
    for relator in gap_relators:
        output.write('%s,\n'%relator)
    output.write('];\n')
    output.write("""
PrintFormatted("{}\\n", info);
start := NanosecondsSinceEpoch();
ans := Length(LowIndexSubgroupsFpGroup(G,%d));
elapsed := Round(Float(NanosecondsSinceEpoch() - start) / 10000000.0)/100;
PrintFormatted("{} subgroups\\n", ans);
PrintFormatted("{} secs\\n", ViewString(elapsed));
"""%ex['index'])

def translate_to_magma(ex, output):
    output.write('"%s; index = %d";\n'%(ex['group'], ex['index']))
    all_relators = ex['short relators'] + ex['long relators']
    relators = [benchmark_util.expand_relator(r) for r in all_relators]
    letters = 'abcdefghijklmnopqrstuvwxyz'
    generators = letters[:ex['rank']]
    output.write('G := Group<\n')
    output.write(', '.join(['%s'%g for g in generators]))
    output.write(' | ')
    output.write(', '.join(['%s'%r for r in relators]))
    output.write('>;\n')
    output.write('T := Time();\n')
    output.write('sgps := LowIndexSubgroups(G, <1, %d>);\n'%ex['index'])
    output.write('T := Time(T);\n')
    output.write('count := #sgps;\n')
    output.write('printf "%o subgroups\\n", count;\n')
    output.write('printf "%o secs\\n", T;\n')

if __name__ == '__main__':
    print(benchmark_util.cpu_info(),
          'with',
          os.cpu_count(),
          'cores (reported by python)/',
          low_index.hardware_concurrency(),
          'cores (reported by C++)', file=sys.stderr)
    if '-gap' in sys.argv:
        with open('/tmp/benchmark.gap', 'w') as gap_script:
            for example in examples:
                if 'gap skip' not in example:
                    translate_to_gap(example, gap_script)
            gap_script.write("QUIT;\n")
        subprocess.run(['sage', '-gap', '/tmp/benchmark.gap'])
        os.unlink('/tmp/benchmark.gap')
    elif '-magma' in sys.argv:
        with open('/tmp/benchmark.magma', 'w') as magma_script:
            for example in examples:
                if 'magma skip' not in example:
                     translate_to_magma(example, magma_script)
            magma_script.write("\nquit;\n")
        subprocess.run(['magma', '/tmp/benchmark.magma'])
        os.unlink('/tmp/benchmark.magma')
    elif '-single' in sys.argv:
        print('Running single-threaded')
        for example in examples:
            run(example, num_threads = 1)
    elif '-regina' in sys.argv:
        import regina
        for example in examples:
            run_regina(example)
    else:
        for example in examples:
            run(example)



import unittest

from collections import Counter

from low_index import *

class TestSimsNode(unittest.TestCase):
    def test_doc(self):
        self.assertIn('A non-abstract SimsNode',
                      SimsNode.__doc__)
        self.assertIn('Create SimsNode for a covering graph ',
                      SimsNode.__init__.__doc__)
        self.assertIn('Copy a SimsNode',
                      SimsNode.__init__.__doc__)

    def test_add_edge(self):
        node = SimsNode(8, 4)
        self.assertEqual(node.rank, 8)

        node.add_edge(2, 1, 1)
        node.add_edge(3, 1, 2)
        node.add_edge(4, 2, 3)
        self.assertTrue(node.verified_add_edge(2, 1, 2))
        self.assertTrue(node.verified_add_edge(2, 2, 3))
        self.assertFalse(node.verified_add_edge(2, 3, 1))

        self.assertFalse(node.is_complete())
        self.assertEqual(node.degree, 3)
        self.assertIn('1--( 3)->2', str(node))

        with self.assertRaises(ValueError):
            node.permutation_rep()

    def test_permutation_rep(self):
        node = SimsNode(2, 2)
        node.add_edge(1, 1, 1)
        node.add_edge(2, 1, 2)
        node.add_edge(2, 2, 1)
        node.add_edge(1, 2, 2)
        self.assertTrue(node.is_complete())
        self.assertEqual(node.degree, 2)
        self.assertEqual(node.permutation_rep(), [[0, 1], [1, 0]])

class TestSimsTree(unittest.TestCase):
    def test_basic(self):
        t = SimsTree(2, 2, [], [])

        perm_reps = [ node.permutation_rep() for node in t.list() ]

        self.assertEqual(
            perm_reps,
            [[[0], [0]],
             [[0, 1], [1, 0]],
             [[1, 0], [0, 1]],
             [[1, 0], [1, 0]]])

    def test_figure_eight(self):
        #                             Manifold("m004").fundamental_group()
        #                     a  a  a  b  A  B  B  A  b
        t = SimsTree(2, 6, [[ 1, 1, 1, 2,-1,-2,-2,-1, 2],
                            [ 2, 1, 1, 1, 2,-1,-2,-2,-1],
                            [-1, 2, 1, 1, 1, 2,-1,-2,-2],
                            [-2,-1, 2, 1, 1, 1, 2,-1,-2],
                            [-2,-2,-1, 2, 1, 1, 1, 2,-1],
                            [-1,-2,-2,-1, 2, 1, 1, 1, 2],
                            [ 2,-1,-2,-2,-1, 2, 1, 1, 1],
                            [ 1, 2,-1,-2,-2,-1, 2, 1, 1],
                            [ 1, 1, 2,-1,-2,-2,-1, 2, 1],
                            [ 1, 1, 1, 2,-1,-2,-2,-1, 2]
                            ],
                     [])

        degrees = Counter([cover.degree for cover in t.list()])
        self.assertEqual(degrees[0], 0)
        self.assertEqual(degrees[1], 1)
        self.assertEqual(degrees[2], 1)
        self.assertEqual(degrees[3], 1)
        self.assertEqual(degrees[4], 2)
        self.assertEqual(degrees[5], 4)
        self.assertEqual(degrees[6], 11)
        self.assertEqual(degrees[7], 0) # Beyond what we counted

class TestPermutationRep(unittest.TestCase):
    def _test_K11n34_7(self, num_threads):
        reps = permutation_reps(
            3,
            ["aaBcbbcAc"],
            ["aacAbCBBaCAAbbcBc"],
            7,
            num_threads = num_threads)

        degrees = Counter([len(rep[0]) for rep in reps])
        self.assertEqual(degrees[1], 1)
        self.assertEqual(degrees[2], 1)
        self.assertEqual(degrees[3], 1)
        self.assertEqual(degrees[4], 1)
        self.assertEqual(degrees[5], 2)
        self.assertEqual(degrees[6], 16)
        self.assertEqual(degrees[7], 30)

        self.assertIn([[0], [0], [0]], reps)
        self.assertIn([[0, 3, 5, 4, 1, 2], [1, 0, 5, 2, 4, 3], [1, 4, 0, 3, 5, 2]], reps)

    def test_K11n34_7_single_threaded(self):
        self._test_K11n34_7(num_threads = 1)

    def test_K11n34_7_multi_threaded(self):
        self._test_K11n34_7(num_threads = 0)

    def test_K11n34_7_fixed_multi_threads(self):
        self._test_K11n34_7(num_threads = 48)

    def test_K15n12345_7(self):
        reps = permutation_reps(
            3,
            ["aBcACAcb"],
            ["aBaCacBAcAbaBabaCAcAbaBaCacBAcAbaBabCAcAbABaCabABAbABaCabCAcAb"],
            7)

        degrees = Counter([len(rep[0]) for rep in reps])
        self.assertEqual(degrees[1], 1)
        self.assertEqual(degrees[2], 1)
        self.assertEqual(degrees[3], 1)
        self.assertEqual(degrees[4], 1)
        self.assertEqual(degrees[5], 3)
        self.assertEqual(degrees[6], 11)
        self.assertEqual(degrees[7], 22)

        self.assertIn([[0], [0], [0]], reps)
        self.assertIn([[0, 2, 1, 3, 4, 6, 5], [0, 3, 4, 5, 1, 2, 6], [1, 2, 0, 5, 6, 3, 4]], reps)
        
    def test_o9_15405_9(self):
        reps = permutation_reps(
            2,
            [],
            ["aaaaabbbaabbbaaaaabbbaabbbaaaaaBBBBBBBB"],
            9)

        degrees = Counter([len(rep[0]) for rep in reps])
        self.assertEqual(degrees[1], 1)
        self.assertEqual(degrees[2], 1)
        self.assertEqual(degrees[3], 1)
        self.assertEqual(degrees[4], 1)
        self.assertEqual(degrees[5], 3)
        self.assertEqual(degrees[6], 3)
        self.assertEqual(degrees[7], 9)
        self.assertEqual(degrees[8], 5)
        self.assertEqual(degrees[9], 14)

        self.assertIn([[0], [0]], reps)
        self.assertIn([[0, 2, 4, 1, 5, 3], [1, 0, 5, 4, 2, 3]], reps)

    def _test_o9_03127_9(self, use_long):
        # Relator is longer than 255 and thus tests that we do not accidentally
        # use a 8-bit integer to index into the relator.
        
        relator = "aabbbaabbaabbbaabbbaaabbbaabbbaabbaabbbaabbbaabbaabbbaabbbaaabbbaabbbaabbaabbbaabbbaaabbbaabbbaabbaabbbaabbbaabbaabbbaabbbaaabbbaabbbaabbaabbbaabbbaaabbbaabbbaabbaabbbaabbbaabbaabbbaabbbaaabbbaabbbaabbaabbbaabbbaabbaabbbaabbbaaabbbaabbbaabbaabbbaabbbaaabbbaabbbaabbaabbb"
        
        reps = permutation_reps(
            2,
            [] if use_long else [ relator ],
            [ relator ] if use_long else [],
            4)

        degrees = Counter([len(rep[0]) for rep in reps])
        
        self.assertEqual(degrees[1], 1)
        self.assertEqual(degrees[2], 3)
        self.assertEqual(degrees[3], 2)
        self.assertEqual(degrees[4], 8)

        self.assertIn([[0, 1, 3, 2], [1, 3, 0, 2]], reps)
        
    def test_o9_03127_9_short(self):
        self._test_o9_03127_9(False)

    def test_o9_03127_9_long(self):
        self._test_o9_03127_9(True)

if __name__ == '__main__':
    print("Number of cores reported by the operating system:",
          hardware_concurrency())
    unittest.main()





import sys
import shutil
from setuptools import setup, Command, Extension

sources = [
    "cpp_src/lowIndex.cpp",
    "cpp_src/words.cpp",
    "cpp_src/coveringSubgraph.cpp",
    "cpp_src/simsNode.cpp",
    "cpp_src/stackedSimsNode.cpp",
    "cpp_src/abstractSimsNode.cpp",
    "cpp_src/simsTreeBase.cpp",
    "cpp_src/simsTree.cpp",
    "cpp_src/simsTreeMultiThreaded.cpp",
    # The pybind11 headers are somewhat heavy - compiling all pieces
    # of the python wrapping in the same translation unit speeds up
    # compilation significantly.
    "cpp_src/wrapAll.cpp"
]

if sys.platform.startswith('win'):
    extra_compile_args = ['/Ox', '/std:c++14', '/MT']
else:
    extra_compile_args = ['-O3', '-std=c++11']

ext_modules = [
    Extension(
        name = 'low_index._low_index',
        sources = sources,
        extra_compile_args = extra_compile_args)
]

class Clean(Command):
    user_options = []
    def initialize_options(self):
        pass
    def finalize_options(self):
        pass
    def run(self):
        for dir in ['build', 'dist', 'low_index.egg-info']:
            shutil.rmtree(dir, ignore_errors=True)
setup(
    packages = ['low_index'],
    package_dir = {'low_index':'python_src'}, 
    ext_modules=ext_modules,
    cmdclass = {'clean':Clean}
)
